library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;

entity co_trace_stream_read_file is
  generic (
    datawidth : positive := 8;
	 inputFilename : string := ""
  );
  port (
    reset, clk : in std_ulogic;
    output_en : out std_ulogic;
    output_rdy : in std_ulogic;
    output_eof : out std_ulogic;
	 output_eos : out std_ulogic;
    output_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end co_trace_stream_read_file;

architecture behavior of co_trace_stream_read_file is
	function hexchar2int(inchar:character) return integer is
				variable x:integer;
		begin
			case inchar is
				when '0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9' => x:=character'pos(inchar)-48;
				when 'A'|'B'|'C'|'D'|'E'|'F' => x:=character'pos(inchar) - 55;
				when 'a'|'b'|'c'|'d'|'e'|'f' => x:=character'pos(inchar) - 87;
				when others => x := 0;--return 0h when invalid character
			end case;
			return x;
		end;

	begin
	
	reader_proc : process
		use std.textio.all;
		variable outDataNibbles : std_ulogic_vector ((((datawidth+7)/8)*8)-1 downto 0);
		variable fileDataBitWidth : integer;
		variable dataNibbleWidth : integer;
		variable data_line : line;
		variable line_cmd : character;
		variable hexChar : character;
		variable hexInt : integer;
		file data_in_file : text;
	begin
		-- init values
		file_open( data_in_file, inputFilename, read_mode );
		output_en 	<= '0';
		output_eof 	<= '0';
		output_eos 	<= '0';

		-- read data bit width from file
		readline(data_in_file, data_line);
		read( data_line, fileDataBitWidth );
		dataNibbleWidth := (fileDataBitWidth+3)/4;
--		if ( dataNibbleWidth < 2 ) then
--			dataNibbleWidth := 2;  -- Bytes are minimum size output, not nibbles
--		end if;
		if ( (dataNibbleWidth/2) /= ((dataNibbleWidth+1)/2) ) then
			dataNibbleWidth := dataNibbleWidth + 1;  -- Bytes are minimum size output, not nibbles always even
		end if;

		-- wait for reset
		wait until rising_edge(clk) and (reset = '0');
		wait until rising_edge(clk);
		
		-- just loop handing out file data 
		while ( not endfile(data_in_file) ) loop
			-- read next data from file.  Line format: <cmd char> 0x<hex data>
			readline(data_in_file, data_line);
			read( data_line, line_cmd );  -- ignored for now, will add detection of close vs. writes later
			if ( line_cmd = 'c' ) then
				output_data <= (others => '0');
				output_eos <= '1';
			else
				output_eos <= '0';
				read( data_line, hexChar );   -- discard ' '
				read( data_line, hexChar );	-- discard '0'
				read( data_line, hexChar );	-- discard 'x'
				for i in dataNibbleWidth downto 1 loop
					read( data_line, hexChar );
					hexInt := hexchar2int( hexChar );
					outDataNibbles((i*4 - 1) downto (i*4 - 4)) := std_ulogic_vector(conv_std_logic_vector(hexInt,4));
				end loop;
				output_data <= outDataNibbles((datawidth-1) downto 0);
			end if;
			output_en 	<= '1';
			wait until rising_edge(clk) and ( output_rdy = '1' );
		end loop;
		
		-- end of file
		file_close( data_in_file );
		output_en 	<= '0';
		output_eof 	<= '1';
		output_eos 	<= '0';
		wait;	-- wait forever
	end process;
end;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;

entity co_trace_stream_write_file is
  generic (
    datawidth : positive := 8;
	 outputFilename : string := "co_trace_stream_write_file_out.txt"
  );
  port (
    reset, clk : in std_ulogic;
    input_en : out std_ulogic;
    input_rdy : in std_ulogic;
    input_eof : in std_ulogic;
	 input_eos : in std_ulogic;
    input_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end co_trace_stream_write_file;

architecture behavior of co_trace_stream_write_file is
	function int2hexchar(inint:integer) return character is
			variable x:character;
		begin
			case inint is
				when 0|1|2|3|4|5|6|7|8|9 => x:=character'val(inint + 48);
				when 10|11|12|13|14|15 => x:=character'val(inint + 55);
				when others => x := 'U';--return 'U' when invalid value
			end case;
			return x;
		end;

	begin
	
	writer_proc : process
		use std.textio.all;
		variable outDataInt : std_ulogic_vector((((datawidth+7)/8)*8)-1 downto 0);
		variable fileDataBitWidth : integer;
		variable dataNibbleWidth : integer;
		variable data_line : line;
		variable line_str : string(1 to 256);
		variable hexChar : character;
		variable hexInt : integer;
		file data_out_file : text;
	begin
		-- init values
		file_open( data_out_file, outputFilename, write_mode );
		input_en 	<= '0';

		-- write data bit width to file
		fileDataBitWidth := datawidth;
		write( data_line, fileDataBitWidth);
		writeline(data_out_file, data_line);

		dataNibbleWidth := outDataInt'length/4;
		if ( dataNibbleWidth < 2 ) then
			dataNibbleWidth := 2;  -- Bytes are minimum size output, not nibbles
		end if;
		-- wait for reset
		wait until rising_edge(clk) and (reset = '0');
		wait until rising_edge(clk);
		
		input_en <= '1';
		-- just loop writing out data to file 
		while ( true ) loop  -- input_eof = '0' ) loop
			wait until rising_edge(clk) and ( input_rdy = '1' );  -- and ( input_eof = '0' );
			if ( input_eof = '1' ) then 
				--exit;
			end if;
			
			if ( input_eos = '1' ) then
				line_str(1) := 'c';
				write(data_line, line_str(1));
			else
				-- write next data to file.  Line format: <cmd char> 0x<hex data>
				line_str(1) := 'w';
				line_str(2) := ' ';
				line_str(3) := '0';
				line_str(4) := 'x';
				outDataInt := std_ulogic_vector(conv_std_logic_vector(0,(outDataInt'length - datawidth))) & input_data;
				for i in dataNibbleWidth downto 1 loop
					hexInt := conv_integer(std_logic_vector(outDataInt((i*4 - 1) downto (i*4 - 4))));
					hexChar := int2hexchar( hexInt );
					line_str(dataNibbleWidth-i+5) := hexChar;
				end loop;
				write( data_line, line_str(1 to (dataNibbleWidth+4)) );
			end if;
			
			writeline(data_out_file, data_line);	
		end loop;
		
		-- end of file
		file_close( data_out_file );
		input_en 	<= '0';
		wait;	-- wait forever
	end process;
end;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;

entity co_trace_stream_monitor_write_file is
  generic (
    datawidth : positive := 8;
	 outputFilename : string := "co_trace_stream_monitor_write_file_out.txt"
  );
  port (
    reset, clk : in std_ulogic;
    input_en : in std_ulogic;
    input_rdy : in std_ulogic;
    input_eof : in std_ulogic;
	 input_eos : in std_ulogic;
    input_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end co_trace_stream_monitor_write_file;

architecture behavior of co_trace_stream_monitor_write_file is
	function int2hexchar(inint:integer) return character is
			variable x:character;
		begin
			case inint is
				when 0|1|2|3|4|5|6|7|8|9 => x:=character'val(inint + 48);
				when 10|11|12|13|14|15 => x:=character'val(inint + 55);
				when others => x := 'U';--return 'U' when invalid value
			end case;
			return x;
		end;

	begin
	
	writer_proc : process
		use std.textio.all;
		variable outDataInt : std_ulogic_vector((((datawidth+7)/8)*8)-1 downto 0);
		variable fileDataBitWidth : integer;
		variable dataNibbleWidth : integer;
		variable data_line : line;
		variable line_str : string(1 to 256);
		variable hexChar : character;
		variable hexInt : integer;
		file data_out_file : text;
	begin
		-- init values
		file_open( data_out_file, outputFilename, write_mode );
		--input_en 	<= '0';

		-- write data bit width to file
		fileDataBitWidth := datawidth;
		write( data_line, fileDataBitWidth);
		writeline(data_out_file, data_line);

		dataNibbleWidth := ((datawidth+7)/8)*2;
		if ( dataNibbleWidth < 2 ) then
			dataNibbleWidth := 2;  -- Bytes are minimum size output, not nibbles
		end if;
		-- wait for reset
		wait until rising_edge(clk) and (reset = '0');
		wait until rising_edge(clk);
		
		--input_en <= '1';
		-- just loop writing out data to file 
		while ( input_eof = '0' ) loop
			wait until rising_edge(clk) and ( input_rdy = '1' ) and ( input_en = '1' ) and ( input_eof = '0' );
			if ( input_eof = '1' ) then 
				exit;
			end if;
			
			-- write next data to file.  Line format: <cmd char> 0x<hex data>
			line_str(1) := 'w';
			line_str(2) := ' ';
			line_str(3) := '0';
			line_str(4) := 'x';
			outDataInt := std_ulogic_vector(conv_std_logic_vector(0,(dataNibbleWidth*4) - datawidth)) & input_data;
			for i in dataNibbleWidth downto 1 loop
				hexInt := conv_integer(std_logic_vector(outDataInt((i*4 - 1) downto (i*4 - 4))));
				hexChar := int2hexchar( hexInt );
				line_str(dataNibbleWidth-i+5) := hexChar;
			end loop;
			
			write( data_line, line_str(1 to (dataNibbleWidth+4)) );
			writeline(data_out_file, data_line);	
		end loop;
		
		-- end of file
		file_close( data_out_file );
		--input_en 	<= '0';
		wait;	-- wait forever
	end process;
end;


