--------------------------------------------------------------------------
-- Copyright (c) 2002-2007 by Impulse Accelerated Technologies, Inc.    --
-- All rights reserved.                                                 --
--                                                                      --
-- This source file may be used and redistributed without charge        --
-- subject to the provisions of the IMPULSE ACCELERATED TECHNOLOGIES,   --
-- INC. REDISTRIBUTABLE IP LICENSE AGREEMENT, and provided that this    --
-- copyright statement is not removed from the file, and that any       --
-- derivative work contains this copyright notice.                      -
--------------------------------------------------------------------------
--                                                                      --
-- avalon_if.vhd: Altera Avalon stream, signal, and DMA interface       --
--   modules                                                            --
--                                                                      --
-- Change History
-- --------------
-- 9/20/2003 - Scott Thibault
--   File created.
-- 12/11/2003 - Scott Thibault
--   Added signal interface.
-- 2/20/2004 - Scott Thibault
--   Added Avalon DMA
-- 4/21/2004 - Scott Thibault
--   Added register interface.
--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_stream is
  generic (
    datawidth : positive := 8
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    stream_en : out std_ulogic;
    stream_rdy : in std_ulogic;
    stream_eos : out std_ulogic;
    stream_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_stream;

architecture rtl of avalon_to_stream is
  signal error, write_addr : std_ulogic;
  signal status : std_logic_vector (31 downto 0);
begin
  write_addr <=
    '1' when avalon_addr = "00" else
    '1' when avalon_addr = "10" else
    '0';

  -- Write to stream
  stream_en <= stream_rdy and avalon_ce and not avalon_write_n and write_addr;
  stream_eos <= '1' when avalon_addr = "10" else '0';
  stream_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_write_n = '0' and write_addr = '1') then
        error <= not stream_rdy;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= stream_rdy;
  status(1) <= '0';
  status(2) <= '0';
  status(3) <= error;
  status(31 downto 4) <= "0000000000000000000000000000";

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity stream_to_avalon is
  generic (
    datawidth : positive := 8
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    stream_en : out std_ulogic;
    stream_rdy : in std_ulogic;
    stream_eos : in std_ulogic;
    stream_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end stream_to_avalon;

architecture rtl of stream_to_avalon is
  signal accept, valid, eos, error, address_stream : std_ulogic;
  signal status : std_logic_vector (31 downto 0);
  signal data : std_ulogic_vector (31 downto 0);
  signal extended_data : std_ulogic_vector (31 downto 0);
begin
  address_stream <= '1' when avalon_addr = "00" else '0';
  -- Read register
  accept <= (not valid or (avalon_ce and not avalon_read_n and address_stream));
  extended_data(31 downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= stream_data;

  read_stream: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      valid <= '0';
      eos <= '0';
    elsif (clk'event and clk='1') then
      if (stream_rdy = '1' and accept = '1') then
        data <= extended_data;
        eos <= stream_eos;
        valid <= '1';
      elsif (avalon_ce = '1' and avalon_read_n = '0' and address_stream = '1') then
        valid <= '0';
      end if;
    end if;
  end process;

  stream_en <= accept;

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_read_n = '0' and address_stream = '1') then
        error <= not valid;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= valid;
  status(1) <= eos;
  status(2) <= '0';
  status(3) <= error;
  status(31 downto 4) <= "0000000000000000000000000000";

  avalon_rdata <=
    std_logic_vector(data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_signal is
  generic (
    datawidth : positive := 32
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    signal_en : out std_ulogic;
    signal_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_signal;

architecture rtl of avalon_to_signal is
  signal write_addr : std_ulogic;
  signal status : std_logic_vector (31 downto 0);
begin
  write_addr <= '1' when avalon_addr = "00" else '0';

  -- Write to signal
  signal_en <= avalon_ce and not avalon_write_n and write_addr;
  signal_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));
 
  -- Status register
  status <= X"00000001"; -- Always ready/no errors.

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity signal_to_avalon is
  generic (
    datawidth : positive := 32
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    signal_en : out std_ulogic;
    signal_rdy : in std_ulogic;
    signal_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end signal_to_avalon;

architecture rtl of signal_to_avalon is
  signal accept, valid, error, address_signal : std_ulogic;
  signal status : std_logic_vector (31 downto 0);
  signal data : std_ulogic_vector (31 downto 0);
  signal extended_data : std_ulogic_vector (31 downto 0);
begin
  address_signal <= '1' when avalon_addr = "00" else '0';
  -- Read register
  accept <= '1';
  extended_data(31 downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= signal_data;

  read_stream: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      valid <= '0';
    elsif (clk'event and clk='1') then
      if (signal_rdy = '1' and accept = '1') then
        data <= extended_data;
        valid <= '1';
      elsif (avalon_ce = '1' and avalon_read_n = '0' and address_signal = '1') then
        valid <= '0';
      end if;
    end if;
  end process;

  signal_en <= accept;

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_read_n = '0' and address_signal = '1') then
        error <= not valid;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= valid;
  status(1) <= '0';
  status(2) <= '0';
  status(3) <= error;
  status(31 downto 4) <= "0000000000000000000000000000";

  avalon_rdata <=
    std_logic_vector(data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_register is
  generic (
    datawidth : positive := 32
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    register_en : out std_ulogic;
    register_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_register;

architecture rtl of avalon_to_register is
  signal write_addr : std_ulogic;
  signal status : std_logic_vector (31 downto 0);
begin
  write_addr <= '1' when avalon_addr = "00" else '0';

  -- Write to signal
  register_en <= avalon_ce and not avalon_write_n and write_addr;
  register_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));
 
  -- Status register
  status <= X"00000001"; -- Always ready/no errors.

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity register_to_avalon is
  generic (
    datawidth : positive := 32
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector (31 downto 0);
    avalon_rdata : out std_logic_vector (31 downto 0);
    register_value : in std_ulogic_vector (datawidth-1 downto 0)
  );
end register_to_avalon;

architecture rtl of register_to_avalon is
  signal extended_data : std_ulogic_vector (31 downto 0);
  signal status : std_logic_vector (31 downto 0);
begin
  extended_data(31 downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= register_value;

  -- Status register
  status <= X"00000001"; -- Always ready/no errors.

  avalon_rdata <=
    std_logic_vector(extended_data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"00000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity avalon_dma is
  port (
    reset, sclk, clk : in std_ulogic;
    AVaddr : out std_logic_vector (31 downto 0);
    AVben : out std_logic_vector (3 downto 0);
    AVread : out std_logic;
    AVreaddata : in std_logic_vector (31 downto 0);
    AVwrite : out std_logic;
    AVwritedata : out std_logic_vector (31 downto 0);
    AVwaitreq : in std_logic;
    ICidata : out std_ulogic_vector (31 downto 0);
    ICaddr : out std_ulogic_vector (31 downto 0);
    ICnextaddr : out std_ulogic_vector (31 downto 0);
    ICwri : out std_ulogic;
    ICodata : in std_ulogic_vector (31 downto 0);
    ICack : out std_ulogic;
    ICreq : in std_ulogic;
    ICmode : in std_ulogic;
    ICbase : in std_ulogic_vector (31 downto 0);
    ICsize : in std_ulogic_vector (2 downto 0);
    ICcount : in std_ulogic_vector (31 downto 0)
  );
end avalon_dma;

architecture rtl of avalon_dma is
  type stateType is (idle,reading,writing);
  signal thisState, nextState : stateType;
  signal address : unsigned(31 downto 0);
  signal counter, counterNext : unsigned(31 downto 0);
  signal done : std_ulogic;
  signal ben : std_logic_vector(3 downto 0);
begin
  -- State machine

  process (reset,clk)
  begin
    if (reset='1') then
      thisState <= idle;
    elsif (clk'event and clk='1') then
      thisState <= nextState;
    end if;
  end process;

  done <= not AVwaitreq when counterNext = unsigned(ICcount) else '0';

  process (thisState,ICreq,ICmode,done)
  begin
    case thisState is
    when idle =>
      if (ICreq='1') then
        if (ICmode='0') then
          nextState <= reading;
        else
          nextState <= writing;
        end if;
      else
        nextState <= idle;
      end if;
    when reading =>
      if (done='1') then
        nextState <= idle;
      else 
        nextState <= reading;
      end if;
    when writing =>
      if (done='1') then
        nextState <= idle;
      else 
        nextState <= writing;
      end if;
    end case;
  end process;

  -- Address generation
  counterNext <= counter + 1;
  process (clk)
  begin
    if (clk'event and clk='1') then
      case thisState is
      when idle =>
        address <= unsigned(ICbase);
        counter <= X"00000000";
      when others =>
        if (AVwaitreq = '0') then
          address <= address + unsigned(ICsize);
          counter <= counterNext;
        end if;
      end case;
    end if;
  end process;

  -- Avalon Master Port signals

  ben <=
    "0001" when ICsize = "001" and address(1 downto 0) = "00" else
    "0010" when ICsize = "001" and address(1 downto 0) = "01" else
    "0100" when ICsize = "001" and address(1 downto 0) = "10" else
    "1000" when ICsize = "001" and address(1 downto 0) = "11" else
    "0011" when ICsize = "010" and address(1) = '0' else
    "1100" when ICsize = "010" and address(1) = '1' else
    "1111";

  AVaddr <= std_logic_vector(address);
  AVben <= ben when thisState = writing else "1111";
  AVread <= '1' when thisState = reading else '0';
  AVwrite <= '1' when thisState = writing else '0';
  AVwritedata <= std_logic_vector(ICodata);

  -- Impulse C component signals

  ICidata <= std_ulogic_vector(AVreaddata);
  ICwri <= not AVwaitreq when thisState = reading else '0';
  ICack <= done when thisState /= idle else '0';
  ICaddr <= std_ulogic_vector(counter);
  ICnextaddr <=
    std_ulogic_vector(counter) when (thisState = idle) or (AVwaitreq = '1') else
    std_ulogic_vector(counterNext);
end rtl;

library ieee;
use ieee.std_logic_1164.all;

package avalon_if is
  component avalon_to_stream is
    generic (
      datawidth : positive := 8
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      stream_en : out std_ulogic;
      stream_rdy : in std_ulogic;
      stream_eos : out std_ulogic;
      stream_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component stream_to_avalon is
    generic (
      datawidth : positive := 8
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      stream_en : out std_ulogic;
      stream_rdy : in std_ulogic;
      stream_eos : in std_ulogic;
      stream_data : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component avalon_to_signal is
    generic (
      datawidth : positive := 32
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      signal_en : out std_ulogic;
      signal_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component signal_to_avalon is
    generic (
      datawidth : positive := 32
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      signal_en : out std_ulogic;
      signal_rdy : in std_ulogic;
      signal_data : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component avalon_to_register is
    generic (
      datawidth : positive := 32
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      register_en : out std_ulogic;
      register_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component register_to_avalon is
    generic (
      datawidth : positive := 32
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector (31 downto 0);
      avalon_rdata : out std_logic_vector (31 downto 0);
      register_value : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component avalon_dma is
    port (
      reset, sclk, clk : in std_ulogic;
      AVaddr : out std_logic_vector (31 downto 0);
      AVben : out std_logic_vector (3 downto 0);
      AVread : out std_logic;
      AVreaddata : in std_logic_vector (31 downto 0);
      AVwrite : out std_logic;
      AVwritedata : out std_logic_vector (31 downto 0);
      AVwaitreq : in std_logic;
      ICidata : out std_ulogic_vector (31 downto 0);
      ICaddr : out std_ulogic_vector (31 downto 0);
      ICnextaddr : out std_ulogic_vector (31 downto 0);
      ICwri : out std_ulogic;
      ICodata : in std_ulogic_vector (31 downto 0);
      ICack : out std_ulogic;
      ICreq : in std_ulogic;
      ICmode : in std_ulogic;
      ICbase : in std_ulogic_vector (31 downto 0);
      ICsize : in std_ulogic_vector (2 downto 0);
      ICcount : in std_ulogic_vector (31 downto 0)
    );
  end component;
end avalon_if;

